from lto.transactions.anchor import Anchor
from lto.transactions.lease import Lease
from lto.transactions.association import Association
from lto.transactions.cancel_lease import CancelLease
from lto.transactions.cancel_sponsorship import CancelSponsorship
from lto.transactions.mass_transfer import MassTransfer
from lto.transactions.revoke_association import RevokeAssociation
from lto.transactions.set_script import SetScript
from lto.transactions.sponsorship import Sponsorship
from lto.transactions.transfer import Transfer
from lto.transactions.register import Register
from lto.transactions.data import Data

def from_data(data):
    if data['type'] == 4:
        return Transfer.from_data(data)
    elif data['type'] == 8:
        return Lease.from_data(data)
    elif data['type'] == 9:
        return CancelLease.from_data(data)
    elif data['type'] == 11:
        return MassTransfer.from_data(data)
    elif data['type'] == 12:
        return Data.from_data(data)
    elif data['type'] == 13:
        return SetScript.from_data(data)
    elif data['type'] == 15:
        return Anchor.from_data(data)
    elif data['type'] == 16:
        return Association.from_data(data)
    elif data['type'] == 17:
        return RevokeAssociation.from_data(data)
    elif data['type'] == 18:
        return Sponsorship.from_data(data)
    elif data['type'] == 19:
        return CancelSponsorship.from_data(data)
    elif data['type'] == 20:
        return Register.from_data(data)
    else:
        raise Exception('Incorrect transaction Type')



import base58
from lto import crypto
from lto.transaction import Transaction
import struct
import math


class Data(Transaction):
    TYPE = 12
    BASE_FEE = 100000000
    VAR_FEE = 10000000
    VAR_BYTES = 256
    DEFAULT_VERSION = 3

    def __init__(self, data):
        super().__init__()

        self.data = self.__dict_to_data(data) if type(data) == dict else data
        self.tx_fee = self.BASE_FEE + math.ceil((len(self.__data_to_binary()) / self.VAR_BYTES)) * self.VAR_FEE
        self.version = self.DEFAULT_VERSION

    @staticmethod
    def __dict_to_data(dictionary):
        data = []
        for key in dictionary:
            data.append(DataEntry.guess(key, dictionary[key]))
        return data

    def __data_to_binary(self):
        binary = b''
        for entry in self.data:
            binary += entry.to_binary()
        return binary

    def __to_binary_v3(self):
        data_binary = self.__data_to_binary()

        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">H", len(self.data)) +
                data_binary)

    def to_binary(self):
        if self.version == 3:
            return self.__to_binary_v3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return (crypto.merge_dicts(
            {
                "type": self.TYPE,
                "version": self.version,
                "sender": self.sender,
                "senderKeyType": self.sender_key_type,
                "senderPublicKey": self.sender_public_key,
                "fee": self.tx_fee,
                "timestamp": self.timestamp,
                "data": list(map(lambda entry: entry.to_json(), self.data)),
                "proofs": self.proofs
            },
            self._sponsor_json()))

    def data_as_dict(self):
        dictionary = {}
        for entry in self.data:
            dictionary[entry.key] = entry.value
        return dictionary

    @staticmethod
    def from_data(data):
        tx = Data([])
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.data = list(map(DataEntry.from_data, data['data'])) if 'data' in data else ''
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''
        return tx


class DataEntry:
    def __init__(self, key, type, value):
        self.key = key
        self.type = type
        self.value = value

    def to_binary(self):
        key_bytes = crypto.str2bytes(self.key)
        return (
                struct.pack(">H", len(key_bytes)) +
                key_bytes +
                self.__value_to_binary()
        )

    def __value_to_binary(self):
        if self.type == 'integer':
            return b'\0' + struct.pack(">Q", self.value)
        elif self.type == 'boolean':
            return b'\1' + (b'\1' if self.value else b'\0')
        elif self.type == 'binary':
            byte_val = crypto.str2bytes(self.value)
            return b'\2' + struct.pack(">H", len(byte_val)) + byte_val
        elif self.type == 'string':
            byte_val = crypto.str2bytes(self.value)
            return b'\3' + struct.pack(">H", len(byte_val)) + byte_val
        else:
            raise Exception('Data Type not supported')

    @staticmethod
    def from_data(data):
        return DataEntry(data['key'], data['type'], data['value'])

    @staticmethod
    def guess(key, value):
        if type(value) == int:
            return DataEntry(key, 'integer', value)
        elif type(value) == bool:
            return DataEntry(key, 'boolean', value)
        elif type(value) == bytes:
            return DataEntry(key, 'binary', value)
        elif type(value) == str:
            return DataEntry(key, 'string', value)
        else:
            raise Exception('Unable to determine type of data entry')

    def to_json(self):
        return {
            "key": self.key,
            "type": self.type,
            "value": self.value
        }



import requests
import json

from lto.transactions import from_data as tx_from_data
from lto.transactions.set_script import SetScript
from lto.accounts.account import Account
from lto import crypto


class PublicNode(object):
    def __init__(self, url, api_key=''):
        self.url = url
        self.api_key = api_key

    def wrapper(self, api, post_data='', host='', headers=None):
        if headers is None:
            headers = {}

        if not host:
            host = self.url

        if self.api_key:
            headers = {"X-API-Key": self.api_key}

        if post_data:
            r = requests.post('%s%s' % (host, api), data=post_data,
                              headers=crypto.merge_dicts(headers, {'content-type': 'application/json'}))
        else:
            r = requests.get('%s%s' % (host, api), headers=headers)

        if r.status_code != 200:
            method = 'POST' if post_data else 'GET'
            json_resp = json.loads(r.text)
            raise Exception(
                '{} {}{} responded with {} {}'.format(method, host, api, r.status_code, r.reason),
                json_resp
            )

        r.raise_for_status()

        return r.json()

    def broadcast(self, transaction):
        data = json.dumps(transaction.to_json())
        response = self.wrapper(api='/transactions/broadcast', post_data=data)
        return tx_from_data(response)

    def compile(self, script_source):
        compiled_script = self.wrapper(api='/utils/script/compile', post_data=script_source)['script']
        return SetScript(compiled_script)

    def height(self):
        return self.wrapper('/blocks/height')['height']

    def last_block(self):
        return self.wrapper('/blocks/last')

    def block(self, n):
        return self.wrapper('/blocks/at/%d' % n)

    def tx(self, id):
        response = self.wrapper('/transactions/info/%s' % id)
        return tx_from_data(response)

    def lease_list(self, address):
        return self.wrapper(api='/leasing/active/{}'.format(address))

    def sponsorship_list(self, address):
        return self.wrapper(api='/sponsorship/status/{}'.format(address))

    def association_list(self, address):
        return self.wrapper(api='/associations/status/{}'.format(address))

    def node_status(self):
        return self.wrapper(api='/node/status')

    def balance(self, address):

        if type(address) == Account:
            address = address.address

        try:
            return self.wrapper('/addresses/balance/%s' % address)['balance']
        except:
            return -1

    def balance_details(self, address):
        if type(address) == Account:
            address = address.address
        return self.wrapper('/addresses/balance/details/%s' % address)


    def transactions(self, limit=100, after='', address=''):
        return self.wrapper('/transactions/address/%s/limit/%d%s' % (
            address, limit, "" if after == "" else "?after={}".format(after)))

    def sign_transaction(self, transaction):
        data = json.dumps(transaction.to_json())
        return(self.wrapper(api='/transactions/sign', post_data=data))




import base58
from lto import crypto
from time import time
from lto.transaction import Transaction
import struct


class Association(Transaction):
    DEFAULT_FEE = 100000000
    TYPE = 16
    DEFAULT_VERSION = 3

    def __init__(self, recipient, association_type, anchor='', expires=0):
        super().__init__()
        self.recipient = recipient
        self.association_type = association_type
        self.anchor = anchor
        self.tx_fee = self.DEFAULT_FEE
        self.version = self.DEFAULT_VERSION

        self.expires = expires
        current = int(time() * 1000)
        if self.expires != 0 and self.expires <= current:
            raise Exception('Wring exipration date')

    def __to_binary_v1(self):
        if self.anchor:
            return (b'\x10' +
                b'\1' +
                crypto.str2bytes(self.chain_id) +
                base58.b58decode(self.sender_public_key) +
                base58.b58decode(self.recipient) +
                struct.pack(">i", self.association_type) +
                b'\1' +
                struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                crypto.str2bytes(self.anchor) +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.tx_fee))
        else:
            return (b'\x10' +
                    b'\1' +
                    crypto.str2bytes(self.chain_id) +
                    base58.b58decode(self.sender_public_key) +
                    base58.b58decode(self.recipient) +
                    struct.pack(">i", self.association_type) +
                    b'\0' +
                    struct.pack(">Q", self.timestamp) +
                    struct.pack(">Q", self.tx_fee))


    def __to_binary_v3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                base58.b58decode(self.recipient) +
                struct.pack(">i", self.association_type) +
                struct.pack(">Q", self.expires) +
                struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                crypto.str2bytes(self.anchor))

    def to_binary(self):
        if self.version == 1:
            return self.__to_binary_v1()
        elif self.version == 3:
            return self.__to_binary_v3()
        else:
            raise Exception('Incorrect Version')



    def to_json(self):
        tx = {
                "type": self.TYPE,
                "version": self.version,
                "sender": self.sender,
                "senderKeyType": self.sender_key_type,
                "senderPublicKey": self.sender_public_key,
                "recipient": self.recipient,
                "associationType": self.association_type,
                "hash": base58.b58encode(crypto.str2bytes(self.anchor)),
                "timestamp": self.timestamp,
                "expires": self.expires if self.version != 1 else None,
                "fee": self.tx_fee,
                "proofs": self.proofs
            }
        if self.version == 1:
            tx.pop('expires')
        return crypto.merge_dicts(tx, self._sponsor_json())


    @staticmethod
    def from_data(data):
        tx = Association(recipient='', association_type='', anchor='')
        tx.type = data['type']
        tx.version = data['version']
        tx.id = data['id'] if 'id' in data else ''
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.recipient = data['recipient']
        tx.association_type = data['associationType']
        tx.hash = data['hash'] if 'hash' in data else ''
        tx.timestamp = data['timestamp']
        tx.expires = data['expires'] if 'expires' in data else ''
        tx.fee = data['fee']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''

        if "sponsor_public_key" in data:
            tx.sponsor = data['sponsor']
            tx.sponsor_public_key = data['sponsorPublicKey']
            tx.sponsor_key_type = data['sponsorKeyType']


        return tx




import base58
from lto import crypto
from lto.accounts.account import Account
from lto.transaction import Transaction
import struct


class Register(Transaction):
    TYPE = 20
    DEFAULT_FEE = 35000000
    DEFAULT_VERSION = 3

    def __init__(self, *accounts):
        super().__init__()

        self.accounts = list(map(self.__account_dict, accounts))
                
        self.tx_fee = self.DEFAULT_FEE
        self.version = self.DEFAULT_VERSION

        if len(self.accounts) > 100:
            raise Exception('Too many accounts')

    @staticmethod
    def __account_dict(account):
      return {'key_type': account.key_type, 'public_key': account.get_public_key()} \
          if isinstance(account, Account) else account

    def __accounts_data(self):
        data = b''
        
        for account in self.accounts:
            data += crypto.key_type_id(account['key_type'])
            data += base58.b58decode(account['public_key'])
        
        return data
    
    def __to_binary_v3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">H", len(self.accounts)) +
                self.__accounts_data())

    def to_binary(self):
        if self.version == 3:
            return self.__to_binary_v3()
        else:
            raise Exception('Incorrect Version ' + self.version)

    @staticmethod
    def __account_to_json(account):
        return {'keyType': account['key_type'], 'publicKey': account['public_key']}

    def to_json(self):
        return (crypto.merge_dicts(
            {
                "type": self.TYPE,
                "version": self.version,
                "sender": self.sender,
                "senderKeyType": self.sender_key_type,
                "senderPublicKey": self.sender_public_key,
                "fee": self.tx_fee,
                "timestamp": self.timestamp,
                "accounts": list(map(self.__account_to_json, self.accounts)),
                "proofs": self.proofs
            },
            self._sponsor_json()))

    @staticmethod
    def __account_from_data(data):
        return {'key_type': data['keyType'], 'public_key': data['publicKey']}

    @staticmethod
    def from_data(data):
        tx = Register()
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.accounts = list(map(Register.__account_from_data, data['accounts']))
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''
        return tx




import base64
import hashlib
import pyblake2
import base58
import inflection
import struct

str2bytes = lambda s: s.encode('latin-1')
bytes2str = lambda b: ''.join(map(chr, b))
str2list = lambda s: [c for c in s]


def sha256(s):
    return hashlib.sha256(str2bytes(s)).digest()


def hash_chain(s):
    a = pyblake2.blake2b(s, digest_size=32).digest()
    b = hashlib.sha256(a).digest()
    return ''.join(map(chr, b))


def get_network(address):
    decoded_address = base58.b58decode(address)
    return str(decoded_address)[6]


def recode(string, from_encoding, to_encoding):
    binary = decode(string, from_encoding)
    return encode(binary, to_encoding)

def decode(string, encoding: str):
    if encoding == 'base58':
        return base58.b58decode(string)
    elif encoding == 'base64':
        return base64.b64decode(string)
    elif encoding == 'hex':
        return bytes.fromhex(string)
    else:
        raise Exception('Failed to decode')


def encode(string, encoding: str):
    if encoding == 'base58':
        return base58.b58encode(string)
    elif encoding == 'base64':
        return base64.b64encode(string)
    elif encoding == 'hex':
        return string.hex()
    else:
        raise Exception('Failed to encode')


def validate_address(address):
    ADDRESS_VERSION = 1
    ADDRESS_CHECKSUM_LENGTH = 4
    ADDRESS_HASH_LENGTH = 20
    ADDRESS_LENGTH = 1 + 1 + ADDRESS_CHECKSUM_LENGTH + ADDRESS_HASH_LENGTH

    addr = bytes2str(base58.b58decode(address))
    if addr[0] != chr(ADDRESS_VERSION):
        raise Exception('Wrong address version')
    elif len(addr) != ADDRESS_LENGTH:
        raise Exception('Wrong address length')
    elif addr[-ADDRESS_CHECKSUM_LENGTH:] != hash_chain(
            str2bytes(addr[:-ADDRESS_CHECKSUM_LENGTH]))[:ADDRESS_CHECKSUM_LENGTH]:
        raise Exception('Wrong address checksum')
    else:
        return True


def key_type_id(key_type):
    if key_type == 'ed25519':
        return b'\1'
    elif key_type == 'secp256k1':
        return b'\2'
    elif key_type == 'secp256r1':
        return b'\3'
    elif key_type == 'rsa':
        return b'\4'
    else:
        raise Exception('Key Type not supported')



def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def compare_data_transaction(data, transaction):
    for key in data:
        key2 = inflection.underscore(key)
        assert data[key] == getattr(transaction, key2)



import base58
import struct
from lto.transaction import Transaction
from lto import crypto


class CancelLease(Transaction):
    TYPE = 9
    DEFAULT_FEE = 100000000
    DEFAULT_VERSION = 3


    def __init__(self, lease_id):
        super().__init__()
        self.lease_id = lease_id
        self.tx_fee = self.DEFAULT_FEE
        self.version = self.DEFAULT_VERSION

    def __to_binary_v2(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\02' +
                crypto.str2bytes(self.chain_id) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">Q", self.timestamp) +
                base58.b58decode(self.lease_id))

    def __to_binary_v3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                base58.b58decode(self.lease_id))

    def to_binary(self):
        if self.version == 2:
            return self.__to_binary_v2()
        elif self.version == 3:
            return self.__to_binary_v3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return(crypto.merge_dicts({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender,
            "senderKeyType": self.sender_key_type,
            "senderPublicKey": self.sender_public_key,
            "fee": self.tx_fee,
            "timestamp": self.timestamp,
            "proofs": self.proofs,
            "leaseId": self.lease_id
        },
        self._sponsor_json()))

    @staticmethod
    def from_data(data):
        tx = CancelLease(lease_id='')
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.lease_id = data['leaseId'] if 'leaseId' in data else ''
        tx.height = data['height'] if 'height' in data else ''

        if "sponsor_public_key" in data:
            tx.sponsor = data['sponsor']
            tx.sponsor_public_key = data['sponsorPublicKey']
            tx.sponsor_key_type = data['sponsorKeyType']

        return tx






from lto.accounts.ed25519.account_ed25519 import AccountED25519 as Account
from lto.accounts.ed25519.account_factory_ed25519 import AccountFactoryED25519 as AccountFactory
from nacl.signing import SigningKey, VerifyKey



from lto.accounts.account import Account
from lto.accounts.account_factory import AccountFactory
from lto.accounts.ecdsa.account_factory_ecdsa import AccountFactoryECDSA
from lto.accounts.ed25519.account_factory_ed25519 import AccountFactoryED25519



from lto.accounts.account import Account
import base58
from lto import crypto


class AccountECDSA(Account):

    def init(self, address, public_key, private_key=None, key_type ='secp256k1',seed=None, nonce=0):
        super().__init__(address, public_key, private_key, key_type, seed, nonce)

    def get_public_key(self):
        return base58.b58encode(self.public_key.to_string(encoding="compressed"))

    def get_private_key(self):
        return base58.b58encode(self.private_key.to_string())

    def sign(self, message):
        if not self.private_key:
            raise Exception("Private key not set")
        return base58.b58encode(self.private_key.sign(message))

    def verify_signature(self, message: str, signature: str, encoding: str = 'base58'):
        if not self.public_key:
            raise Exception('Unable to verify message; no public sign key')
        raw_signature = crypto.decode(signature, encoding)
        return self.public_key.verify(raw_signature, message)



import os
from lto import crypto
from lto.word_list import wordList

def random_seed():
    word_count = len(wordList)
    words = []
    for i in range(5):
        r = crypto.bytes2str(os.urandom(4))
        x = (ord(r[3])) + (ord(r[2]) << 8) + (ord(r[1]) << 16) + (ord(r[0]) << 24)
        w1 = x % word_count
        w2 = ((int(x / word_count) >> 0) + w1) % word_count
        w3 = ((int((int(x / word_count) >> 0) / word_count) >> 0) + w2) % word_count
        words.append(wordList[w1])
        words.append(wordList[w2])
        words.append(wordList[w3])
    return ' '.join(words)



from __future__ import absolute_import, division, print_function, unicode_literals

import base58

from lto.accounts import AccountFactory
from lto.public_node import PublicNode
from lto.accounts import AccountFactoryECDSA, AccountFactoryED25519


class LTO:

    def __init__(self, chain_id='T'):
        if chain_id == 'T':
            self.NODE = PublicNode('https://testnet.lto.network')
        elif chain_id == 'L':
            self.NODE = PublicNode('https://nodes.lto.network')
        else:
            self.NODE = ''

        self.chain_id = chain_id
        self.account_factories = {
            'ed25519': AccountFactoryED25519(chain_id),
            'secp256r1': AccountFactoryECDSA(chain_id, curve='secp256r1'),
            'secp256k1': AccountFactoryECDSA(chain_id, curve='secp256k1')
        }

    def Account(self, public_key=None, private_key=None, key_type='ed25519', seed=None, seed_method=None, nonce=0):
        factory = self.account_factories[key_type].with_seed_method(seed_method)

        if seed:
            account = factory.create_from_seed(seed, nonce)
        elif private_key:
            account = factory.create_from_private_key(private_key)
        elif public_key:
            account = factory.create_from_public_key(public_key)
        else:
            account = factory.create()

        # assert public_key is None or account.get_public_key() == self.__key_base58(public_key), "Public key mismatch"
        # assert private_key is None or account.get_private_key() == self.__key_base58(private_key), "Private key mismatch"

        return account

    @staticmethod
    def __key_base58(key):
        if type(key) == object:
            return base58.b58encode(bytes(key))
        elif type(key) == bytes:
            return base58.b58encode(key)
        else:
            return key



import copy
from abc import ABC, abstractmethod
from lto.accounts.brainwallet import random_seed as brainwallet_random_seed
from lto.accounts.bip39 import random_seed as bip39_random_seed


class AccountFactory(ABC):
    def __init__(self, chain_id, seed_method):
        self.chain_id = chain_id
        self.seed_method = seed_method

    def with_seed_method(self, seed_method):
        if seed_method is None or self.seed_method == seed_method:
            return self

        clone = copy.copy(self)
        clone.seed_method = seed_method
        return clone

    @abstractmethod
    def create_sign_keys(self, seed, nonce):
        pass

    @abstractmethod
    def create_address(self, public_key):
        pass

    def create(self):
        if self.seed_method == 'brainwallet':
            seed = brainwallet_random_seed()
        else:
            raise Exception('Seed method under construction')

        '''elif self.seed_method == 'bip39':
            seed = bip39_random_seed()
        elif self.seed_method.startswith('bip39:'):
            # raise Exception("Method under construction")
            seed = bip39_random_seed(self.seed_method[6:])
        else:
            raise Exception('Unknown seed method')'''

        return self.create_from_seed(seed)

    @abstractmethod
    def create_from_seed(self, seed, nonce=0):
        pass

    @abstractmethod
    def create_from_private_key(self, private_key):
        pass

    @abstractmethod
    def create_from_public_key(self, public_key):
        pass

    @abstractmethod
    def create_with_values(self, address, public_key, private_key, key_type, seed=''):
        pass

